/*! \file copyright.h
 *  \brief Copyright and license information.

 ATHENA 

 AUTHORS:
     James M. Stone, Princeton Univeristy
     Thomas A. Gardiner, Sandia National Laboratory
     Peter J. Teuben, University of Maryland
     John F. Hawley, University of Virginia

     This software was produced with support from NSF Grant AST-0113571 

 Copyright 2003 James M. Stone, Thomas A. Gardiner,
                Peter J. Teuben, John F. Hawley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 */
